@@ -36,6 +36,7 @@ import numpy as np
 import tensorflow.compat.v2 as tf
 
 from deepconsensus.utils import dc_constants
+from deepconsensus.utils import utils
 
 
 # Define base fields for TFRecords.
@@ -46,7 +47,9 @@ PROTO_FEATURES_INFERENCE = {
     # Shapes are written to the int64_list of the example.
     'subreads/shape': tf.io.FixedLenFeature(shape=[3], dtype=tf.int64),
     'subreads/num_passes': tf.io.FixedLenFeature(shape=[1], dtype=tf.int64),
+    'ccs_base_quality_scores': tf.io.FixedLenFeature(shape=[], dtype=tf.int64),
 }
+
 # Add on label fields to train proto.
 PROTO_FEATURES_TRAIN = dict(
     {
@@ -56,16 +59,25 @@ PROTO_FEATURES_TRAIN = dict(
 
 
 def get_total_rows(max_passes: int) -> int:
-  """Returns total rows in input tf.Examples. Update if other signals added."""
-  # For each of `max_subreads`, we have three pieces of information: bases, PW,
-  # and IP. We also have four rows for SN, and one for strand.
-  # The information is structured as follows:
-  # Bases: (0, params.max_passes - 1) represent bases.
-  # PW: rows params.max_passes to (params.max_passes * 2 - 1)
-  # IP: rows (params.max_passes * 2) to (params.max_passes * 3 - 1)
-  # Strand: rows (params.max_passes * 3) to (params.max_passes * 4)
-  # CCS+SN: rows (params.max_passes * 4 + 1) to (params.max_passes * 4 + 5)
-  # The last five rows are CCS sequence (1), and SN (4).
+  """Returns total rows in input tf.Examples.
+
+  Update if other signals added.
+
+  For each of `max_subreads`, we have four pieces of information: bases, PW, IP,
+  and strand. We also have one row for CCS, and four rows for SN (in that
+  order).
+  The information is structured as follows:
+  Bases: rows 0 to  (params.max_passes - 1)
+  PW: rows (params.max_passes) to (params.max_passes * 2 - 1)
+  IP: rows (params.max_passes * 2) to (params.max_passes * 3 - 1)
+  Strand: rows (params.max_passes * 3) to (params.max_passes * 4 - 1)
+  CCS+SN: rows (params.max_passes * 4) to (params.max_passes * 4 + 5)
+
+  Args:
+    max_passes: Maximum number of subreads to show. Space is made for them all
+      even though few examples will have enough subreads to fill these rows.
+  Returns: Total number of rows in the full example.
+  """
   return (max_passes * 4) + 5
 
 
@@ -93,14 +105,15 @@ def remove_internal_gaps_and_shift(label: tf.Tensor) -> tf.Tensor:
 
 def format_rows(
     subreads: tf.Tensor,
-    params: ml_collections.FrozenConfigDict,
+    max_passes: int,
+    max_length: int,
     cap_pw: bool = True,
     cap_ip: bool = True,
     cap_sn: bool = True,
 ) -> tf.Tensor:
   """Returns model input matrix formatted based on input args."""
   base_indices, pw_indices, ip_indices, strand_indices, ccs_indices, sn_indices = get_indices(
-      params.max_passes)
+      max_passes)
   base_rows = subreads[slice(*base_indices)]
   pw_rows = subreads[slice(*pw_indices)]
   ip_rows = subreads[slice(*ip_indices)]
@@ -117,31 +130,20 @@ def format_rows(
   if cap_sn:
     sn_rows = tf.clip_by_value(
         sn_rows, clip_value_min=0, clip_value_max=dc_constants.SN_MAX)
-  if params.get('input_format') == 'stack[base,pw,ip,sn]':
-    ccs_rows = tf.image.pad_to_bounding_box(ccs_rows, 0, 0, params.max_passes,
-                                            params.max_length)
-    sn_rows = tf.image.pad_to_bounding_box(sn_rows, 0, 0, params.max_passes,
-                                           params.max_length)
-    rows = tf.concat([base_rows, pw_rows, ip_rows, ccs_rows, sn_rows], axis=-1)
-    if params.max_passes < 32:
-      rows = tf.image.pad_to_bounding_box(rows, 0, 0, 32, params.max_length)
-    num_rows = max(params.max_passes, 32)
-  else:
-    rows = tf.concat(
-        [base_rows, pw_rows, ip_rows, strand_rows, ccs_rows, sn_rows], axis=0)
-    num_rows = get_total_rows(params.max_passes)
-
-  rows.set_shape((num_rows, params.max_length, params.num_channels))
+  rows = tf.concat(
+      [base_rows, pw_rows, ip_rows, strand_rows, ccs_rows, sn_rows], axis=0)
+  num_rows = get_total_rows(max_passes)
+  rows.set_shape((num_rows, max_length, 1))
   return rows
 
 
 def process_feature_dict(
-    features: Dict[str, tf.Tensor],
+    features: Dict[str, Union[np.ndarray, int, bytes]],
     params: Union[config_dict.ConfigDict, config_dict.FrozenConfigDict],
     cap_pw: bool = True,
     cap_ip: bool = True,
     cap_sn: bool = True,
-) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]:
+) -> Dict[str, Union[np.ndarray, int, bytes]]:
   """Parses a serialized tf.Example to return an input, label, and metadata.
 
   Args:
@@ -155,24 +157,43 @@ def process_feature_dict(
     rows: Input matrix that will be fed into neural networks for training.
     label: Label vector that will be used for training.
     num_passes: The number of subreads present in this example.
+    window_position: The position at which this example starts within the ccs
+        read.
+    name: Name of the ZMW, e.g. "m64011_181218_235052/315/ccs".
   """
   label = tf.convert_to_tensor(np.array([]))
   subreads = features['subreads']
   num_passes = features['subreads/num_passes']
   rows = format_rows(
       subreads=subreads,
-      params=params,
+      max_passes=params.max_passes,
+      max_length=params.max_length,
       cap_pw=cap_pw,
       cap_ip=cap_ip,
       cap_sn=cap_sn)
-  return rows, label, num_passes, features['window_pos'], features['name']
+  # Don't forget to update DC_FEATURES in dc_constants.py if new features are
+  # added/removed.
+  features = {
+      'rows': rows,
+      'label': label,
+      'num_passes': num_passes,
+      'window_pos': features['window_pos'],
+      'name': features['name'],
+      'ccs_base_quality_scores': features['ccs_base_quality_scores']
+  }
+  return features
 
 
-def parse_example(proto_string, inference=False):
+def parse_example(proto_string: Dict[str, tf.Tensor],
+                  inference: bool = False,
+                  max_length: int = 120) -> Dict[str, tf.Tensor]:
+  """Parses serialized Training or Inference TF.Examples."""
   if inference:
     proto_features = PROTO_FEATURES_INFERENCE
   else:
     proto_features = PROTO_FEATURES_TRAIN
+  if not proto_features['ccs_base_quality_scores'].shape:
+    proto_features['ccs_base_quality_scores'].shape.append(max_length)
   parsed_features = tf.io.parse_single_example(
       serialized=proto_string, features=proto_features)
   return parsed_features
@@ -185,7 +206,7 @@ def process_input(
     cap_pw: bool = True,
     cap_ip: bool = True,
     cap_sn: bool = True,
-) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
+) -> Dict[str, tf.Tensor]:
   """Parses a serialized tf.Example to return an input, label, and metadata.
 
   Args:
@@ -200,17 +221,20 @@ def process_input(
     rows: Input matrix that will be fed into neural networks for training.
     label: Label vector that will be used for training.
     num_passes: The number of subreads present in this example.
+    window_position: The position at which this example starts within the ccs
+        read.
+    name: Name of the ZMW, e.g. "m64011_181218_235052/315/ccs".
   """
-  parsed_features = parse_example(proto_string, inference)
-  flat_subreads = tf.io.decode_raw(parsed_features['subreads/encoded'],
+  features = parse_example(proto_string, inference, params.max_length)
+  flat_subreads = tf.io.decode_raw(features['subreads/encoded'],
                                    dc_constants.TF_DATA_TYPE)
-  subreads = tf.reshape(flat_subreads, parsed_features['subreads/shape'])
-  num_passes = tf.cast(parsed_features['subreads/num_passes'],
+  subreads = tf.reshape(flat_subreads, features['subreads/shape'])
+  num_passes = tf.cast(features['subreads/num_passes'],
                        dc_constants.TF_DATA_TYPE)
   if not inference:
-    flat_label = tf.io.decode_raw(parsed_features['label/encoded'],
+    flat_label = tf.io.decode_raw(features['label/encoded'],
                                   dc_constants.TF_DATA_TYPE)
-    label = tf.reshape(flat_label, parsed_features['label/shape'])
+    label = tf.reshape(flat_label, features['label/shape'])
 
     if params.remove_label_gaps:
       label = remove_internal_gaps_and_shift(label)
@@ -219,18 +243,36 @@ def process_input(
     label = tf.convert_to_tensor(np.array([]))
   rows = format_rows(
       subreads=subreads,
-      params=params,
+      max_passes=params.max_passes,
+      max_length=params.max_length,
       cap_pw=cap_pw,
       cap_ip=cap_ip,
       cap_sn=cap_sn)
-  return rows, label, num_passes
+  rows = {
+      'rows': rows,
+      'label': label,
+      'num_passes': num_passes,
+      'window_pos': features['window_pos'],
+      'name': features['name'],
+      'ccs_base_quality_scores': features['ccs_base_quality_scores'],
+  }
+  return rows
 
 
-def drop_metadata(*args):
+def tf_example_to_training_tuple(
+    tf_example: Dict[str, tf.Tensor]) -> Tuple[tf.Tensor, tf.Tensor]:
   """Return only subreads and label."""
-  subreads = args[0]
-  label = args[1]
-  return (subreads, label)
+  return (tf_example['rows'], tf_example['label'])
+
+
+def filter_ccs_q_score(tf_example: Dict[str, tf.Tensor],
+                       max_phred_qual: int = 0) -> bool:
+  """Returns True if phred is less than max_phred_qual."""
+  # When max_phred_qual is 0, do not filter.
+  if max_phred_qual == 0:
+    return True
+  phred = utils.tf_avg_phred(tf_example['ccs_base_quality_scores'])
+  return phred <= max_phred_qual
 
 
 def get_dataset(file_pattern: str,
@@ -243,13 +285,33 @@ def get_dataset(file_pattern: str,
                 cap_ip: bool = True,
                 limit: int = -1,
                 drop_remainder: bool = True,
-                keep_metadata: bool = False) -> tf.data.Dataset:
-  """Parses TFRecords and return a dataset."""
+                example_label_tuple: bool = False) -> tf.data.Dataset:
+  """Parses TFRecords and return a dataset.
 
-  # Output type annotations added for clarity, but Pytype only expects
-  # Tuple[Any, Any] here and does not check for tf.Tensors.
-  def _process_input_helper(
-      proto_string: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
+  Args:
+    file_pattern: File path(s) to be parsed by create_glob_list.
+    num_epochs: How many epochs for which to repeat.
+    batch_size: How many examples should be in each batch.
+    params: Hyperparameters used to format the subreads into rows.
+    inference: Whether to parse tf.Examples for inference or training.
+    cap_pw: If True, pulse width values are capped.
+    cap_ip: If True, interpulse distance values are capped.
+    limit: Max number of examples to get. Set to -1 for no limit.
+    drop_remainder: Passed to TFRecordDataset.batch
+    example_label_tuple: If True, output simplified format for training/eval
+      as (rows, label)
+
+  Returns:
+    A dataset for which each batch has the following elements:
+    rows: Input matrix that will be fed into neural networks for training.
+    label: Label vector that will be used for training.
+    num_passes: The number of subreads present in this example.
+    window_position: The position at which this example starts within the ccs
+        read.
+    name: Name of the ZMW, e.g. "m64011_181218_235052/315/ccs".
+  """
+
+  def _process_input_helper(proto_string: tf.Tensor) -> Dict[str, tf.Tensor]:
     return process_input(
         proto_string=proto_string,
         params=params,
@@ -257,17 +319,27 @@ def get_dataset(file_pattern: str,
         cap_ip=cap_ip,
         inference=inference)
 
+  def _filter_q_helper(tf_example: Dict[str, tf.Tensor]) -> bool:
+    return filter_ccs_q_score(tf_example, params.max_phred_qual)
+
   file_patterns = create_glob_list(file_pattern)
   ds = tf.data.TFRecordDataset(file_patterns, compression_type='GZIP')
   ds = ds.map(map_func=_process_input_helper)
+
+  if params.max_phred_qual:
+    ds = ds.filter(_filter_q_helper)
+
   ds = ds.shuffle(buffer_size=params.buffer_size, reshuffle_each_iteration=True)
+
   if num_epochs:
     ds = ds.repeat(num_epochs)
 
-  # When training, we can drop num_passes, window_pos, and name.
-  if not keep_metadata:
+  # When training, we can drop num_passes, window_position, and name.
+  if example_label_tuple:
     ds = ds.map(
-        drop_metadata, num_parallel_calls=tf.data.AUTOTUNE, deterministic=False)
+        tf_example_to_training_tuple,
+        num_parallel_calls=tf.data.AUTOTUNE,
+        deterministic=False)
   ds = ds.batch(batch_size=batch_size, drop_remainder=drop_remainder)
   ds = ds.prefetch(tf.data.experimental.AUTOTUNE)
   ds = ds.take(limit)
@@ -287,8 +359,7 @@ def create_glob_list(paths: Union[str, List[str]]) -> List[str]:
 def create_input_fn(params, mode, limit: int = -1, drop_remainder: bool = True):
   """Returns an input function that will return a tfrecord based dataset."""
 
-  def _process_input_helper(
-      proto_string: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
+  def _process_input_helper(proto_string: tf.Tensor) -> Dict[str, tf.Tensor]:
     # Set inference to False here because we only use this function with
     # tf.Examples that have labels present.
     return process_input(
@@ -298,6 +369,9 @@ def create_input_fn(params, mode, limit: int = -1, drop_remainder: bool = True):
         cap_pw=True,
         cap_ip=True)
 
+  def _filter_q_helper(tf_example: Dict[str, tf.Tensor]) -> bool:
+    return filter_ccs_q_score(tf_example, params.max_phred_qual)
+
   def input_fn() -> tf.data.Dataset:
     """Prepares a dataset for training or evaluation."""
     is_training = (mode == 'train')
@@ -310,9 +384,6 @@ def create_input_fn(params, mode, limit: int = -1, drop_remainder: bool = True):
         lambda x: tf.data.TFRecordDataset(x, compression_type='GZIP'),
         num_parallel_calls=tf.data.AUTOTUNE,
         deterministic=False)
-    if is_training:
-      ds = ds.shuffle(
-          buffer_size=params['buffer_size'], reshuffle_each_iteration=True)
 
     # Best practices suggest batching first, but this map errors out when we
     # batch first, so do this before mapping.
@@ -320,10 +391,19 @@ def create_input_fn(params, mode, limit: int = -1, drop_remainder: bool = True):
         _process_input_helper,
         num_parallel_calls=tf.data.experimental.AUTOTUNE,
         deterministic=False)
+
+    ds = ds.filter(_filter_q_helper, 'ccs_quality_filter')
+
+    if is_training:
+      ds = ds.shuffle(
+          buffer_size=params['buffer_size'], reshuffle_each_iteration=True)
+
     # Best practices suggest batching before mapping.
     ds = ds.batch(batch_size, drop_remainder=drop_remainder)
     ds = ds.map(
-        drop_metadata, num_parallel_calls=tf.data.AUTOTUNE, deterministic=False)
+        tf_example_to_training_tuple,
+        num_parallel_calls=tf.data.AUTOTUNE,
+        deterministic=False)
     ds = ds.prefetch(tf.data.experimental.AUTOTUNE)
     ds = ds.take(limit)
     return ds
