@@ -28,8 +28,8 @@
 r"""Inference binary for all neural network models.
 
 To use this binary for running inference with a specific model, the
-corresponding config does not need to be specified and will be inferred. Example
-usage:
+corresponding config does not need to be specified and will be inferred.
+Example usage:
 
 OUT_DIR=/tmp
 CHECKPOINT=<internal>
@@ -49,6 +49,7 @@ import ml_collections
 from ml_collections.config_flags import config_flags
 import tensorflow as tf
 
+from deepconsensus.models import data_providers
 from deepconsensus.models import model_utils
 
 FLAGS = flags.FLAGS
@@ -82,18 +83,21 @@ def run_inference(out_dir: str, params: ml_collections.ConfigDict,
 
   with strategy.scope():
     model = model_utils.get_model(params)
+    row_size = data_providers.get_total_rows(params.max_passes)
+    input_shape = (1, row_size, params.max_length, params.num_channels)
+    model_utils.print_model_summary(model, input_shape)
     checkpoint = tf.train.Checkpoint(model=model)
-    checkpoint.restore(checkpoint_path)
+    # Need to run a forward pass with the model in order for
+    # assert_existing_objects_matched to work as expected. If you don't do this,
+    # then assert_existing_objects_matched will not raise an error even if the
+    # wrong checkpoint is used. Some context here: b/148023980
+    checkpoint.restore(checkpoint_path).assert_existing_objects_matched()
 
     model.compile(
         optimizer=tf.keras.optimizers.Adam(learning_rate=params.learning_rate),
         loss=model_utils.get_deepconsensus_loss(params),
         metrics=model_utils.get_deepconsensus_metrics())
 
-    input_shape = (1, params.hidden_size, params.max_length,
-                   params.num_channels)
-    model_utils.print_model_summary(model, input_shape)
-
     model_utils.run_inference_and_write_results(
         model=model, out_dir=out_dir, params=params, limit=limit)
 
