@@ -27,10 +27,11 @@
 # SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 """Utilities for running training and inference."""
 
+import io
 import json
 import logging
 import os
-from typing import List, Optional, Tuple
+from typing import Any, Dict, List, Optional, Tuple, Union
 
 import ml_collections
 import numpy as np
@@ -39,8 +40,8 @@ import tensorflow as tf
 from deepconsensus.models import data_providers
 from deepconsensus.models import losses_and_metrics
 from deepconsensus.models import networks
+from deepconsensus.models import transformer_basic_params
 from deepconsensus.utils import dc_constants
-from official.nlp.transformer import misc
 
 
 def get_deepconsensus_loss(
@@ -74,6 +75,8 @@ def get_deepconsensus_metrics(name_prefix='') -> List[tf.keras.metrics.Metric]:
       tf.keras.metrics.SparseCategoricalAccuracy(name=f'{name_prefix}accuracy'),
       losses_and_metrics.PerExampleAccuracy(
           name=f'{name_prefix}per_example_accuracy'),
+      losses_and_metrics.AlignmentMetric(
+          name=f'{name_prefix}alignment_identity')
   ] + per_class_accuracy_metrics
 
 
@@ -114,11 +117,9 @@ def get_model(params: ml_collections.ConfigDict) -> tf.keras.Model:
   """Returns desired model based on the given params."""
   if params.model_name == 'fc':
     model = networks.FullyConnectedNet(params)
-  elif params.model_name == 'conv_net':
-    model = networks.ConvNet(params)
   elif params.model_name == 'transformer':
     model = networks.EncoderOnlyTransformer(params)
-  elif params.model_name == 'transformer_learn_values':
+  elif 'transformer_learn_values' in params.model_name:
     model = networks.EncoderOnlyLearnedValuesTransformer(params)
   else:
     raise ValueError('Unknown model name: %s' % params.model_name)
@@ -203,7 +204,7 @@ def modify_params(params: ml_collections.ConfigDict,
     else:
       params.max_length = extract_max_length(params.train_path)
 
-    if params.model_name == 'transformer_learn_values':
+    if 'transformer_learn_values' in params.model_name:
       dim = ((params.use_bases * params.per_base_hidden_size) +
              (params.use_pw * params.pw_hidden_size) +
              (params.use_ip * params.ip_hidden_size) +
@@ -218,21 +219,19 @@ def modify_params(params: ml_collections.ConfigDict,
       params.hidden_size += 1
 
     # Set model-specific parameters
-    if params.model_name == 'conv_net':
-      # If max_passes < 32; image will be padded as to a height of(32) req'd.
-      params.hidden_size = max(32, params.max_passes)
-    elif params.model_name == 'transformer':
+    if params.model_name == 'transformer':
       # Transformer code uses default_batch_size, whereas my code uses
       # batch_size, so make sure both are the same.
       params.default_batch_size = params.batch_size
-    elif params.model_name == 'transformer_learn_values':
+    elif 'transformer_learn_values' in params.model_name:
       # Transformer code uses default_batch_size, whereas my code uses
       # batch_size, so make sure both are the same.
       params.default_batch_size = params.batch_size
       if params.condense_transformer_input:
+        logging.info('Setting hidden size to transformer_input_size.')
         params.hidden_size = params.transformer_input_size
     if 'transformer' in params.model_name:
-      transformer_params = misc.get_model_params(
+      transformer_params = get_transformer_model_params(
           params.transformer_model_size, num_gpus=num_gpus)
       # Only add hyperparameters that don't already exist.
       for param_name, param_value in transformer_params.items():
@@ -240,6 +239,25 @@ def modify_params(params: ml_collections.ConfigDict,
           params[param_name] = param_value
 
 
+def get_transformer_model_params(param_set, num_gpus):
+  """Gets predefined transformer model params."""
+  params_map = {
+      'tiny': transformer_basic_params.TINY_PARAMS,
+      'base': transformer_basic_params.BASE_PARAMS,
+      'big': transformer_basic_params.BIG_PARAMS,
+  }
+  if num_gpus > 1:
+    if param_set == 'big':
+      return transformer_basic_params.BIG_MULTI_GPU_PARAMS.copy()
+    elif param_set == 'base':
+      return transformer_basic_params.BASE_MULTI_GPU_PARAMS.copy()
+    else:
+      raise ValueError('Not valid params: param_set={} num_gpus={}'.format(
+          param_set, num_gpus))
+
+  return params_map[param_set].copy()
+
+
 def run_inference_and_write_results(model: tf.keras.Model,
                                     out_dir: str,
                                     params: ml_collections.ConfigDict,
@@ -265,7 +283,8 @@ def run_inference_and_write_results(model: tf.keras.Model,
           drop_remainder=False,
           # `inference` is set to False because this function is only used with
           # tf.Examples formatted for training mode as they contain a label.
-          inference=False)
+          inference=False,
+          example_label_tuple=True)
 
       history = model.evaluate(
           x=validation_dataset, batch_size=params.batch_size, steps=None)
@@ -292,7 +311,12 @@ def print_model_summary(model: tf.keras.Model, input_shape: Tuple[int, int, int,
 
 def read_params_from_json(checkpoint_path: str) -> ml_collections.ConfigDict:
   """Reads the params read from the params.json file for given checkpoint."""
-  json_path = os.path.join(os.path.dirname(checkpoint_path), 'params.json')
+  # For SavedModel, the path could be to the directory itself.
+  if os.path.isdir(checkpoint_path):
+    dir_path = checkpoint_path
+  else:
+    dir_path = os.path.dirname(checkpoint_path)
+  json_path = os.path.join(dir_path, 'params.json')
   params = ml_collections.ConfigDict(
       json.load(tf.io.gfile.GFile(json_path, 'r')))
   # This cannot be passed in to be updated as json serializes it as a string.
@@ -302,3 +326,124 @@ def read_params_from_json(checkpoint_path: str) -> ml_collections.ConfigDict:
     del params.dtype
     params.dtype = dc_constants.TF_DATA_TYPE
   return params
+
+
+class DTypeEncoder(json.JSONEncoder):
+  """json encoder that allows for dtypes to be encoded."""
+
+  def default(self, o: Any) -> Any:
+    if isinstance(o, tf.DType):
+      return repr(o)
+    else:
+      return json.JSONEncoder.default(self, o)
+
+
+def save_params_as_json(out_dir: str,
+                        params: ml_collections.ConfigDict) -> None:
+  """Saves params to a JSON file."""
+  json_path = os.path.join(out_dir, 'params.json')
+  tf.io.gfile.makedirs(os.path.dirname(json_path))
+  with tf.io.gfile.GFile(json_path, 'w') as json_file:
+    json_file.write(json.dumps(dict(params), indent=4, cls=DTypeEncoder))
+
+
+def get_datasets(
+    params: ml_collections.ConfigDict, strategy: tf.distribute.Strategy
+) -> Tuple[tf.distribute.DistributedDataset, tf.distribute.DistributedDataset]:
+  """Returns datasets for training and evaluation."""
+  train_input_fn = data_providers.create_input_fn(
+      params=params, mode='train', limit=params.limit)
+  eval_input_fn = data_providers.create_input_fn(
+      params=params, mode='eval', limit=params.limit)
+  train_dataset = strategy.experimental_distribute_dataset(train_input_fn())
+  eval_dataset = strategy.experimental_distribute_dataset(eval_input_fn())
+  return train_dataset, eval_dataset
+
+
+def get_step_counts(params: ml_collections.ConfigDict,
+                    eval_and_log_every_step: bool) -> Tuple[int, int]:
+  """Returns the steps for training and evaluation."""
+
+  if eval_and_log_every_step:
+    steps_per_epoch = 1
+    steps_per_eval = 1
+  elif params.limit <= 0:
+    steps_per_epoch = params.n_examples_train // params.batch_size
+    steps_per_eval = params.n_examples_eval // params.batch_size
+  else:
+    # When `params.limit` is set, use it to determine epoch size.
+    steps_per_epoch = max(1, params.limit // params.batch_size)
+    steps_per_eval = max(1, params.limit // params.batch_size)
+  return steps_per_epoch, steps_per_eval
+
+
+def get_checkpoint_and_initial_epoch(
+    model: tf.keras.models.Model, optimizer: tf.keras.optimizers.Optimizer,
+    out_dir: str, steps_per_epoch: int) -> Tuple[tf.train.Checkpoint, int]:
+  """Loads a checkpoint if available and sets epoch to start training."""
+  checkpoint = tf.train.Checkpoint(model=model, optimizer=optimizer)
+  latest_checkpoint = tf.train.latest_checkpoint(out_dir)
+  initial_epoch = 0
+  if latest_checkpoint:
+    checkpoint.restore(latest_checkpoint)
+    logging.info('Loaded checkpoint %s', latest_checkpoint)
+    initial_epoch = optimizer.iterations.numpy() // steps_per_epoch
+  return checkpoint, initial_epoch
+
+
+def reset_all_metrics(metrics: List[tf.keras.metrics.Metric]) -> None:
+  """Resets the values of provided metrics."""
+  for metric in metrics:
+    metric.reset_states()
+
+
+def log_and_save_metrics(epoch: int, step: int, total_steps: int,
+                         optimizer: tf.keras.optimizers.Optimizer,
+                         losses_dict: Dict[str, float],
+                         metrics: List[tf.keras.metrics.Metric],
+                         training: bool) -> None:
+  """Logs metrics and saves them for TensorBoard."""
+  logging.info(
+      'epoch: %d  step: %d of %d metrics: %s', epoch, step, total_steps,
+      ' '.join(f'{loss_name}= {losses_dict[loss_name]}'
+               for loss_name in losses_dict.keys()))
+
+  if training:
+    tf.summary.scalar('learning_rate', optimizer.lr, step=optimizer.iterations)
+  for loss_name in losses_dict.keys():
+    tf.summary.scalar(
+        loss_name, losses_dict[loss_name], step=optimizer.iterations)
+  for metric in metrics:
+    tf.summary.scalar(metric.name, metric.result(), step=optimizer.iterations)
+
+
+def write_row(handle: Union[io.TextIOWrapper], row: List[Any]) -> None:
+  """Formats an array as tab-delimited and writes."""
+  handle.write('\t'.join(map(str, row)) + '\n')
+
+
+def save_checkpoint(checkpoint: tf.train.Checkpoint, out_dir: str,
+                    train_metrics: List[tf.keras.metrics.Metric],
+                    eval_metrics: List[tf.keras.metrics.Metric],
+                    write_checkpoint_metrics: bool) -> str:
+  """Save checkpoint and return its name."""
+  checkpoint_name = checkpoint.save(os.path.join(out_dir, 'checkpoint'))
+  logging.info('Saved checkpoint to %s', checkpoint_name)
+  logging.info('Logging checkpoint %s metrics.', checkpoint_name)
+  metrics_file = os.path.join(out_dir, 'checkpoint_metrics.tsv')
+  if write_checkpoint_metrics:
+    if not tf.io.gfile.exists(metrics_file):
+      with tf.io.gfile.GFile(metrics_file, 'w') as f:
+        row = ['checkpoint_name', 'group', 'name', 'value']
+        write_row(f, row)
+
+    with tf.io.gfile.GFile(metrics_file, 'a') as f:
+      for group_name, metrics in [('train', train_metrics),
+                                  ('eval', eval_metrics)]:
+        for metric in metrics:
+          row = [
+              checkpoint_name, group_name, metric.name,
+              float(metric.result())
+          ]
+          write_row(f, row)
+  return checkpoint_name
